from abc import ABC, abstractmethod


class PipelineConfig:

    def __init__(self, pipeline_name, pipeline_id, config, pipeline_package, pipeline_class, target_problem):
        self.pipeline_name = pipeline_name
        self.pipeline_id = pipeline_id
        self.config = config
        self.pipeline_package = pipeline_package
        self.pipeline_class = pipeline_class
        self.target_problem = target_problem

    def to_dict(self):
        return {
            'pipeline_name': self.pipeline_name,
            'pipeline_id': self.pipeline_id,
            'config': self.config,
            'pipeline_package': self.pipeline_package,
            'pipeline_class': self.pipeline_class,
            'target_problem': self.target_problem
        }

    @staticmethod
    def from_dict(config_dict: dict):
        return PipelineConfig(
            pipeline_name=config_dict['pipeline_name'],
            pipeline_id=config_dict['pipeline_id'],
            config=config_dict['config'],
            pipeline_package=config_dict['pipeline_package'],
            pipeline_class=config_dict['pipeline_class'],
            target_problem=config_dict['target_problem']
        )

    def __str__(self):
        return self.to_dict()

    def _repr_html_(self):
        return """
        <p><h3>%s</h3></p>
        <div><b>id</b>: %s</div>
        <div><b>target_problem</b>: %s</div>
        <div><b>package</b>: %s</div>
        <div><b>class</b>: %s</div>
        <div><b>config</b>: %s</div>
        """ % (self.pipeline_name, self.pipeline_id, self.target_problem,
               self.pipeline_package, self.pipeline_class, self.config)


class BasePipeline(ABC):

    @abstractmethod
    def save_pipeline(self, pipeline_id) -> PipelineConfig:
        """
        trigger inside models to saving
        :return: config dict
        """
        pass

    @staticmethod
    @abstractmethod
    def load_pipeline(config, model_manager):
        """
        :param pipeline_id:
        :return:
        """
        pass

    @abstractmethod
    def pipeline_name(self) -> str:
        """
        :return: pipelines name
        """
        pass

    @abstractmethod
    def predict(self, features):
        pass

    @abstractmethod
    def predict_proba(self, features):
        pass

    @abstractmethod
    def pipeline_score(self):
        pass


from abc import ABC, abstractmethod
from functools import lru_cache
from uuid import uuid4

from spt_datascience.datascience.singleton import Singleton
from spt_datascience.datascience.model_manager import ModelManager
from spt_datascience.datascience.pipelines.base_pipeline import BasePipeline, PipelineConfig
from importlib import import_module


class PipelineManager(metaclass=Singleton):

    @abstractmethod
    def get_pipeline(self, pipeline_id) -> BasePipeline:
        """
        Get pipelines by pipeline_id
        :param pipeline_id:
        :return: pipelines object
        """
        pass

    @abstractmethod
    def save_pipeline(self, pipeline: BasePipeline) -> str:
        """
        Save pipelines info
        :param pipeline: pipelines to save
        :return:
        """
        pass


class SPTPipelineManager(PipelineManager):

    def __init__(self, spt_resource_factory, model_manager):
        self.spt_resource_factory = spt_resource_factory
        self.model_manager = model_manager

    def load_pipeline_object(self, pipeline_package, pipeline_class, config, model_manager):
        module = import_module(pipeline_package)
        return getattr(
            module, pipeline_class
        ).load_pipeline(config, model_manager)

    @lru_cache(maxsize=512)
    def get_pipeline(self, pipeline_id) -> BasePipeline:
        with self.spt_resource_factory.get_mongo() as mongo_client:
            pipeline_config_dict = mongo_client.spt.pipelines.find_one({'pipeline_id': pipeline_id})
        pipeline_config = PipelineConfig.from_dict(pipeline_config_dict)

        pipeline_package = pipeline_config.pipeline_package
        pipeline_class = pipeline_config.pipeline_class

        return self.load_pipeline_object(
            pipeline_package,
            pipeline_class,
            pipeline_config,
            self.model_manager
        )

    def save_pipeline(self, pipeline: BasePipeline) -> str:
        pipeline_id = str(uuid4())
        pipeline_config = pipeline.save_pipeline(pipeline_id).to_dict()
        with self.spt_resource_factory.get_mongo() as mongo_client:
            mongo_client.spt.pipelines.insert_one(pipeline_config)
        return pipeline_id

    def save_pipeline_config(self, pipeline_config: PipelineConfig):
        with self.spt_resource_factory.get_mongo() as mongo_client:
            mongo_client.spt.pipelines.insert_one(pipeline_config.to_dict())

    def delete_pipeline(self, pipeline_id):
        with self.spt_resource_factory.get_mongo() as mongo_client:
            mongo_client.spt.pipelines.delete_one({'id': pipeline_id})





from abc import ABC, abstractmethod


class Credentials(ABC):
    __slots__ = 'crede', 'crede_object', 'factory_params', 'custom_params', 'spt_resource_factory'

    def __init__(self, spt_resource_factory, spt_ds_factory, factory_params, custom_params):
        self.spt_resource_factory = spt_resource_factory
        self.spt_ds_factory = spt_ds_factory
        self.factory_params = factory_params
        self.custom_params = custom_params

    def get_credentials(self):
        return self.crede


class ModelManagerCredentials(Credentials):

    def __init__(self, spt_resource_factory, spt_ds_factory, factory_params, custom_params):
        super().__init__(spt_resource_factory, spt_ds_factory, factory_params, custom_params)
        self.crede = dict(spt_resource_factory=spt_resource_factory, spt_ds_factory=spt_ds_factory)


class PipelineManagerCredentials(Credentials):

    def __init__(self, spt_resource_factory, spt_ds_factory, factory_params, custom_params):
        super().__init__(spt_resource_factory, spt_ds_factory, factory_params, custom_params)
        self.crede = dict(spt_resource_factory=spt_resource_factory, spt_ds_factory=spt_ds_factory)


class S3ModelManagerCredentials(Credentials):

    def __init__(self, spt_resource_factory, spt_ds_factory, factory_params, custom_params):
        super().__init__(spt_resource_factory, spt_ds_factory, factory_params, custom_params)
        self.crede = dict(spt_resource_factory=spt_resource_factory, spt_ds_factory=spt_ds_factory)


class MLFlowCredentials(Credentials):

    def __init__(self, spt_resource_factory, spt_ds_factory, factory_params, custom_params):
        super().__init__(spt_resource_factory, spt_ds_factory, factory_params, custom_params)
        with spt_resource_factory.get_mongo() as mongo_client:
            mlflow_crede = mongo_client.spt.credentials.find_one({"type": "mlflow"})
        self.crede = mlflow_crede



from abc import ABC, abstractmethod


class ModelConfig:

    def __init__(self, id, name, version, extra, model_package, model_class, bins, storage_type='s3'):
        self.id = id
        self.name = name
        self.storage_type = storage_type
        self.extra = extra
        self.version = version
        self.model_package = model_package
        self.model_class = model_class
        self.bins = bins

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'version': self.version,
            'extra': self.extra,
            'model_package': self.model_package,
            'model_class': self.model_class,
            'bins': self.bins,
            'storage_type': self.storage_type
        }

    @staticmethod
    def from_dict(config_dict):

        return ModelConfig(
            id=config_dict['id'],
            name=config_dict['name'],
            version=config_dict['version'],
            extra=config_dict['extra'],
            model_package=config_dict['model_package'],
            model_class=config_dict['model_class'],
            bins=config_dict['bins'],
            storage_type=config_dict.get('storage_type', 's3')
        )

    def __str__(self):
        return self.to_dict()

    def _repr_html_(self):
        return """
        <p><h3>%s</h3></p>
        <div><b>id</b>: %s</div>
        <div><b>version</b>: %s</div>
        <div><b>package</b>: %s</div>
        <div><b>class</b>: %s</div>
        <div><b>extra</b>: %s</div>
        """ % (self.name, self.id, self.version, self.model_package, self.model_class, self.extra)


class BaseModel(ABC):

    @abstractmethod
    def save_model(self, model_id, version) -> ModelConfig:
        """
        Save model files
        :return: config dict
        """
        raise NotImplemented()

    @staticmethod
    @abstractmethod
    def load_model(config: ModelConfig):
        """

        :param config: dict with config
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def model_name(self) -> str:
        """
        :return: model name
        """
        raise NotImplemented()



from psycopg2.extensions import register_adapter, AsIs
from functools import partial
import numpy as np


def addapt_numpy_float64(numpy_float64):
    return AsIs(numpy_float64)


def addapt_numpy_int64(numpy_int64):
    return AsIs(numpy_int64)


register_adapter(np.float64, addapt_numpy_float64)
register_adapter(np.int64, addapt_numpy_int64)


def _insert_batch_df(df, conn, table):
    cursor = conn.cursor()
    cols = df.columns.tolist()
    values = [cursor.mogrify("(%s)" % ','.join('%s' for _ in cols), tup).decode('utf8') for tup in
              df.itertuples(index=False)]
    query = "INSERT INTO %s(%s) VALUES " % (table, ','.join(cols)) + ",".join(values)
    cursor.execute(query)
    conn.commit()
    cursor.close()


def insert_df(df, conn, table, batch_size=1_000_000):
    size = df.shape[0]
    for ind in range(0, size, batch_size):
        _insert_batch_df(df.iloc[ind:ind + batch_size], conn, table)


class DataVault:

    @staticmethod
    def hash_cols(cols, cur):
        """
        This method returns sql injection helping to make a hash of values
        """
        hashed_cols = ','.join(map(
            lambda col: f"coalesce(nullif(upper(trim(cast(%s as varchar))), ''), '^^')" %
                        cur.mogrify("%s", (col,)).decode('utf-8'),
            cols
        ))
        return f"cast(md5(nullif(concat_ws('||',%s), '^^||^^')) as text)" % hashed_cols

    @staticmethod
    def hash_col(col, cur):
        return f"cast(md5(nullif(concat_ws('||',%s), '^^||^^')) as text)" % \
               f"coalesce(nullif(upper(trim(cast(%s as varchar))), ''), '^^')" % \
               cur.mogrify("%s", (col,)).decode('utf-8')

    @staticmethod
    def hub(data_df, config, conn):
        """
        Use it to upload data to hub of Data Vault
        :param data_df: pandas dataframe
        :param config: config file
        :param conn: connection to database
        """
        schema = config.get('schema', 'business_vault')
        src_pk = config.get('src_pk')
        src_nk = config.get('src_nk')
        destination = schema + '.' + config['table']
        record_source = config.get('record_source')

        with conn.cursor() as cur:
            hash_col = partial(DataVault.hash_col, cur=cur)
            data_df[src_pk] = data_df[src_nk].apply(hash_col)
            args_str = ','.join(
                "(%s, " % row[src_pk] + cur.mogrify(
                    "%s, %s, now())", (row[src_nk], record_source)
                ).decode('utf-8')
                for i, row in data_df.iterrows()
            )
            query = "insert into %s (%s) VALUES " % (
                destination, ','.join([src_pk, src_nk, 'record_source', 'load_date'])) + args_str
            cur.execute(query)

    @staticmethod
    def satellite(data_df, config, conn):
        """
        Use it to upload data to satellite of Data Vault
        :param data_df: pandas dataframe
        :param config: config file
        :param conn: connection to database
        """
        schema = config.get('schema', 'business_vault')
        src_pk = config.get('src_pk')
        src_pk_computable = config.get('src_pk_computable')
        src_nk = config.get('src_nk')
        src_hashdiff = config.get('src_hashdiff')
        destination = schema + '.' + config['table']
        src_payload = config.get('src_payload')
        record_source = config.get('record_source')

        with conn.cursor() as cur:
            hash_col = partial(DataVault.hash_col, cur=cur)
            hash_cols = partial(DataVault.hash_cols, cur=cur)
            if src_nk is not None:
                if isinstance(src_nk, list):
                    data_df[src_pk] = data_df[src_nk].apply(hash_cols)
                else:
                    data_df[src_pk] = data_df[src_nk].apply(hash_col)
            elif src_pk_computable is not None:
                data_df[src_pk] = data_df[src_pk_computable]
            else:
                data_df[src_pk] = data_df[src_pk].apply(
                    lambda pk: cur.mogrify('%s', (pk,)).decode('utf-8')
                )
            data_df['hashdiff'] = data_df[src_payload].apply(hash_cols, axis=1)
            args_str = ','.join(
                "(%s, " % row[src_pk] + cur.mogrify(
                    "%s, now(), now()," + "%s," * len(src_payload), [record_source] + row[src_payload].tolist()
                ).decode('utf-8') + row['hashdiff'] + ")"
                for i, row in data_df.iterrows()
            )
            query = "insert into %s (%s) VALUES " % (
                destination, ','.join([src_pk, 'record_source', 'effective_from', 'load_date'] + src_payload + [src_hashdiff])) \
                    + args_str
            cur.execute(query)

    @staticmethod
    def link(data_df, config, conn):
        """
        Use it to upload data to link of Data Vault
        :param data_df: pandas dataframe
        :param config: config file
        :param conn: connection to database
        """
        schema = config.get('schema', 'business_vault')
        src_pk = config.get('src_pk')
        src_fk = config.get('src_fk')
        l_link, r_link = src_fk[0], src_fk[1]
        destination = schema + '.' + config['table']
        data_df['record_source'] = config.get('record_source')

        with conn.cursor() as cur:
            hash_cols = partial(DataVault.hash_cols, cur=cur)
            data_df['link_pk'] = data_df[src_fk].apply(hash_cols, axis=1)
            args_str = ','.join("(%s, " % row['link_pk'] + cur.mogrify(
                "%s, %s, %s, now())", (row[l_link], row[r_link], row['record_source'])
            ).decode('utf-8')
                                for i, row in data_df.iterrows())
            query = "insert into %s (%s) VALUES " % (
                destination, ','.join([src_pk, l_link, r_link, 'record_source', 'load_date'])) + args_str
            cur.execute(query)



    @staticmethod
    def multi_link(data_df, config, conn):
        """
        Use it to upload data to link of Data Vault
        :param data_df: pandas dataframe
        :param config: config file
        :param conn: connection to database
        """
        schema = config.get('schema', 'business_vault')
        src_pk = config.get('src_pk')
        src_fk = config.get('src_fk')
        nk_fk_map = config.get('nk_fk_map')
        destination = schema + '.' + config['table']
        data_df['record_source'] = config.get('record_source')

        with conn.cursor() as cur:
            hash_col = partial(DataVault.hash_col, cur=cur)
            hash_cols = partial(DataVault.hash_cols, cur=cur)

            for fk in src_fk:
                if fk not in nk_fk_map.values():
                    data_df[fk] = data_df[fk].apply(
                        lambda fk: cur.mogrify('%s', (fk,)).decode('utf-8')
                    )

            for nk, fk in nk_fk_map.items():
                data_df[fk] = data_df[nk].apply(hash_col)
            data_df[src_pk] = data_df[src_fk].apply(hash_cols, axis=1)

            def build_value(row):
                return "(%s, " % row[src_pk] + \
                    ', '.join(row[fk] for fk in src_fk) + \
                    cur.mogrify(", %s, now())", (row['record_source'],)).decode('utf-8')

            args_str = ','.join(
                    build_value(row)
            for i, row in data_df.iterrows())

            query = "insert into %s (%s) VALUES %s" % (
                destination, ','.join([
                    src_pk,
                    *(fk for fk in src_fk),
                    'record_source', 'load_date']),
                args_str
            )
            cur.execute(query)



from abc import ABC, abstractmethod

import mlflow
import os

from spt_datascience.credentials import Credentials
from spt_datascience.datascience.model_manager import SPTModelManager
from spt_datascience.datascience.model_util import MongoModelUtil
from spt_datascience.datascience.models_storage import  S3ModelStorage
from spt_datascience.datascience.pipeline_manager import SPTPipelineManager



class Resource(ABC):

    def __init__(self, c: Credentials):
        self.c = c

    @abstractmethod
    def get_object(self):
        pass

    @staticmethod
    @abstractmethod
    def get_name():
        pass


class ModelManagerResource(Resource):

    def get_object(self):
        credential = self.c.get_credentials()
        return SPTModelManager(
            spt_resource_factory=credential['spt_resource_factory'],
            spt_ds_factory=credential['spt_ds_factory']
        )

    @staticmethod
    def get_name():
        return 'model_manager'


class MLFlowResource(Resource):

    def get_object(self):
        credential = self.c.get_credentials()
        os.environ['MLFLOW_TRACKING_USERNAME'] = credential['user']
        os.environ['MLFLOW_TRACKING_PASSWORD'] = credential['password']
        mlflow.set_tracking_uri(credential['tracking_uri'])
        return mlflow

    @staticmethod
    def get_name():
        return 'mlflow'


class PipelineManagerResource(Resource):

    def get_object(self):
        credential = self.c.get_credentials()
        return SPTPipelineManager(
            spt_resource_factory=credential['spt_resource_factory'],
            model_manager=credential['spt_ds_factory'].get_model_manager()
        )

    @staticmethod
    def get_name():
        return 'pipeline_manager'



from abc import abstractmethod

from spt_datascience.datascience.singleton import Singleton
from spt_datascience.datascience.models.base_model import ModelConfig
from datetime import datetime


class ModelUtil(metaclass=Singleton):

    @abstractmethod
    def get_model_version(self, model_name) -> int:
        """
        return last model version
        :param model_name: name of the model
        :return: model version
        """
        pass

    @abstractmethod
    def save_model_config(self, model_config: ModelConfig):
        """
        :param model_config: model info
        """
        pass


class MongoModelUtil(ModelUtil):

    def __init__(self, spt_resource_factory):
        self.spt_resource_factory = spt_resource_factory

    @staticmethod
    def _fix_dict(d: dict):
        return {
            str(key): value if type(value) != dict else MongoModelUtil._fix_dict(value)
            for key, value in d.items()
        }

    def get_model_version(self, model_name) -> int:
        with self.spt_resource_factory.get_mongo() as mongo_client:
            models_info = list(mongo_client.spt.models.find({"name": model_name}).sort("version", -1))
        for model_info in models_info:
            return model_info['version']
        return 0

    def save_model_config(self, model_config):
        model_config_dict = model_config.to_dict()
        model_config_dict['date'] = str(datetime.now())
        with self.spt_resource_factory.get_mongo() as mongo_client:
            mongo_client.spt.models.insert_one(
                MongoModelUtil._fix_dict(model_config_dict)
            )



from .factory import DsFactory
from .datascience.data_vault import DataVault
from .datascience.models.base_model import ModelConfig, BaseModel
from .datascience.models.estimator_model import EstimatorModel
from .datascience.pipelines.base_pipeline import PipelineConfig, BasePipeline


from spt_datascience import BaseModel, ModelConfig


class EstimatorModel(BaseModel):

    def __init__(self, estimator):
        self.estimator = estimator

    def save_model(self, model_id, version) -> ModelConfig:
        raise NotImplemented("Save mlflow model using mlflow API")

    @staticmethod
    def load_model(config: ModelConfig):
        return EstimatorModel(
            estimator=config.bins['estimator']
        )

    def model_name(self) -> str:
        return 'mlflow_s'


from abc import ABC, abstractmethod
from importlib import import_module

from spt_datascience.datascience.models.base_model import BaseModel, ModelConfig
from spt_datascience.datascience.singleton import Singleton

MODELS_BUCKET = 'theme-models'


class ModelStorage():

    @abstractmethod
    def upload_model_bins(self, model_config):
        pass

    @abstractmethod
    def load_model_bins(self, model_config):
        pass


class S3ModelStorage(ModelStorage, metaclass=Singleton):

    def __init__(self, spt_resource_factory):
        self.s3_client = spt_resource_factory.get_s3_manager()
        self.mongo_client = spt_resource_factory.get_mongo()

    def upload_model_bins(self, model_config):
        for bin_name, bin_obj in model_config.bins.items():
            model_path = model_config.id + '/' + bin_name
            self.s3_client.upload_bin(bucket_name=MODELS_BUCKET, id=model_path, bin_str=bin_obj)
            model_config.bins[bin_name] = model_path
        return model_config

    def load_model_bins(self, model_config):
        model_bins = model_config.bins
        model_bucket = MODELS_BUCKET
        for bin_name, bin_path in model_bins.items():
            model_bins[bin_name] = self.s3_client.download_bin(bucket_name=model_bucket, id=bin_path)
        model_config.bins = model_bins
        return model_config

    def delete_model(self, model_id):
        self.mongo_client.spt.models.delete_one({'id': model_id})
        self.s3_client.delete_folder(MODELS_BUCKET, model_id)


class MLFlowSklearnModelStorage(ModelStorage, metaclass=Singleton):

    def __init__(self, spt_resource_factory, ds_factory):
        self.spt_resource_factory = spt_resource_factory
        self.ds_factory = ds_factory
        self.mlflow = ds_factory.get_mlflow()

    def upload_model_bins(self, model_config):
        raise NotImplemented("Save mlflow model using mlflow API")

    def load_model_bins(self, model_config):
        model_bins = model_config.bins
        for bin_name, bin_value in model_bins.items():
            model_uri = f"models:/{bin_value['name']}/{bin_value['version']}"
            model_bins[bin_name] = self.mlflow.sklearn.load_model(model_uri)
        return model_config




#!/usr/bin/env python
import sys
from setuptools import find_namespace_packages, setup, find_packages

package_name = "spt_datascience"
package_version = "0.0.8.dev1"
description = """SPT ds manager"""


if sys.version_info < (3, 7):
    print("Error: spt-factory does not support this version of Python.")
    print("Please upgrade to Python 3.7 or higher.")
    sys.exit(1)


setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=description,
    author='Mark Poroshin',
    author_email='mporoshin@smartpredictiontech.ru',
    packages=find_namespace_packages(include=[
        'spt_datascience',
        'spt_datascience.*'
    ]),
    # url="https://gitlab.com/dv_group/dv_elt_lib",
    include_package_data=True,
    install_requires=[
        "psycopg2-binary==2.9.3",
        "pymongo==4.1.1",
        "boto3==1.24.45",
        "ulid-py==1.1.0",
        "numpy"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
      ],
)



from abc import ABC, abstractmethod
from spt_datascience.resources import ModelManagerResource, PipelineManagerResource, MLFlowResource
from spt_datascience.credentials import ModelManagerCredentials, PipelineManagerCredentials, MLFlowCredentials
from functools import partial


class Factory(ABC):

    @abstractmethod
    def get_resource_crede_pairs(self):
        pass

    def __init__(self, spt_resource_factory, **factory_params):
        self.factory_params = factory_params
        self.spt_resource_factory = spt_resource_factory
        self.resources = {}
        for resource, credentials in self.get_resource_crede_pairs():

            def get_resource(resource, credentials, **params):
                c = credentials(
                    spt_resource_factory=self.spt_resource_factory,
                    spt_ds_factory=self,
                    factory_params=self.factory_params,
                    custom_params=params
                )
                return resource(c).get_object()

            def get_credentials(credentials, **params):
                return credentials(
                    spt=self,
                    factory_params=self.factory_params,
                    custom_params=params
                ).get_credentials()

            setattr(self, f'get_{resource.get_name()}', partial(get_resource, resource=resource, credentials=credentials))
            setattr(self, f'get_{resource.get_name()}_credentials', partial(get_credentials, credentials=credentials))


class DsFactory(Factory):

    def get_resource_crede_pairs(self):
        return (
            (ModelManagerResource, ModelManagerCredentials),
            (PipelineManagerResource, PipelineManagerCredentials),
            (MLFlowResource, MLFlowCredentials)
        )



from abc import abstractmethod
from importlib import import_module

import ulid

from spt_datascience.datascience.model_util import ModelUtil, MongoModelUtil
from spt_datascience.datascience.models.base_model import BaseModel, ModelConfig
from functools import lru_cache
from spt_datascience.datascience.singleton import Singleton
from spt_datascience.datascience.models_storage import S3ModelStorage
from spt_datascience.datascience.models_storage import MLFlowSklearnModelStorage


class ModelManager(metaclass=Singleton):

    @abstractmethod
    def get_model(self, model_id) -> BaseModel:
        """
        Return model object by id
        :param model_id: id of the model
        :return:
        """
        pass

    @abstractmethod
    def save_model(self, model, storage_type) -> str:
        """
        Save model and return model id
        :param model: model inherit from BaseModel
        :return: model id
        """
        pass

    @abstractmethod
    def get_model_config(self, model_id, storage_type) -> ModelConfig:
        """
        Return model object by id
        :param model_id: id of the model
        :return:
        """
        pass

    @abstractmethod
    def save_model_config(self, model_config, storage_type):
        """
        Save model config
        :param model: model inherit from BaseModel
        """
        pass


class SPTModelManager(ModelManager):

    def __init__(self, spt_resource_factory, spt_ds_factory):
        self.ds_factory = spt_ds_factory
        self.model_util = MongoModelUtil(spt_resource_factory)
        self.model_storages = {
            's3': S3ModelStorage(spt_resource_factory),
            'mlflow/sklearn': MLFlowSklearnModelStorage(spt_resource_factory, spt_ds_factory)
        }
        self.spt_resource_factory = spt_resource_factory

    @lru_cache(maxsize=1024)
    def get_model(self, model_id) -> BaseModel:
        model_config = self.get_model_config(model_id)
        model_package = model_config.model_package
        model_class = model_config.model_class
        return self._load_model_object(model_package, model_class, model_config)

    def save_model(self, model: BaseModel, storage_type: str = 's3') -> str:
        model_name = model.model_name()
        model_version = self.increment_model_version(self.model_util.get_model_version(model_name))
        model_id = self.produce_model_id(model_name, model_version)
        model_config = model.save_model(model_id, model_version)
        self.save_model_config(model_config, storage_type)
        return model_id

    def produce_model_id(self, model_name, model_version):
        return f"{model_name}{model_version}#{ulid.new().str}"

    def increment_model_version(self, version):
        return version + 1

    def get_model_config(self, model_id, storage_type: str = 's3'):
        with self.spt_resource_factory.get_mongo() as mongo_client:
            model_config_dict = mongo_client.spt.models.find_one({'id': model_id})
        storage_type = model_config_dict.get('storage_type', storage_type)
        model_config = ModelConfig.from_dict(model_config_dict)
        return self.model_storages[storage_type].load_model_bins(model_config)

    def save_model_config(self, model_config, storage_type: str = 's3'):
        storage_type = model_config.storage_type if model_config.storage_type else storage_type
        model_config = self.model_storages[storage_type].upload_model_bins(model_config)
        self.model_util.save_model_config(model_config)

    def delete_model(self, model_id, storage_type: str = 's3'):
        self.model_storages[storage_type].delete_model(model_id)

    def _load_model_object(self, model_package, model_class, model_config):
        module = import_module(model_package)
        return getattr(
            module, model_class
        ).load_model(model_config)
